# -*- coding: utf-8 -*-
"""
Implements higher mathematical functions.

The functions are described in :doc:`Mellin space </theory/Mellin>`.
"""

import numba as nb
import numpy as np


@nb.njit(cache=True)
def cern_polygamma(Z, K):  # pylint: disable=all
    """
    Computes the polygamma functions :math:`\\psi_k(z)`.

    Reimplementation of ``WPSIPG`` (C317) in `CERNlib <http://cernlib.web.cern.ch/cernlib/>`_
    :cite:`KOLBIG1972221`.

    Note that the SciPy implementation :data:`scipy.special.digamma`
    does not allow for complex inputs.

    Parameters
    ----------
        Z : complex
            argument of polygamma function
        K : int
            order of polygamma function

    Returns
    -------
        H : complex
            k-th polygamma function :math:`\\psi_k(z)`
    """
    # fmt: off
    DELTA = 5e-13
    R1 = 1
    HF = R1/2
    C1 = np.pi**2
    C2 = 2*np.pi**3
    C3 = 2*np.pi**4
    C4 = 8*np.pi**5

    # SGN is originally indexed 0:4 -> no shift
    SGN = [-1,+1,-1,+1,-1]
    # FCT is originally indexed -1:4 -> shift +1
    FCT = [0,1,1,2,6,24]

    # C is originally indexed 1:6 x 0:4 -> swap indices and shift new last -1
    C = nb.typed.List()
    C.append([
            8.33333333333333333e-2,
           -8.33333333333333333e-3,
            3.96825396825396825e-3,
           -4.16666666666666667e-3,
            7.57575757575757576e-3,
           -2.10927960927960928e-2])
    C.append([
            1.66666666666666667e-1,
           -3.33333333333333333e-2,
            2.38095238095238095e-2,
           -3.33333333333333333e-2,
            7.57575757575757576e-2,
           -2.53113553113553114e-1])
    C.append([
            5.00000000000000000e-1,
           -1.66666666666666667e-1,
            1.66666666666666667e-1,
           -3.00000000000000000e-1,
            8.33333333333333333e-1,
           -3.29047619047619048e+0])
    C.append([
            2.00000000000000000e+0,
           -1.00000000000000000e+0,
            1.33333333333333333e+0,
           -3.00000000000000000e+0,
            1.00000000000000000e+1,
           -4.60666666666666667e+1])
    C.append([10., -7., 12., -33., 130., -691.])
    U=Z
    X=np.real(U)
    A=np.abs(X)
    if K < 0 or K > 4:
        raise NotImplementedError("Order K has to be in [0:4]")
    if np.abs(np.imag(U)) < DELTA and np.abs(X+int(A)) < DELTA:
        raise ValueError("Argument Z equals non-positive integer")
    K1=K+1
    if X < 0:
        U=-U
    V=U
    H=0
    if A < 15:
        H=1/V**K1
        for I in range(1,14-int(A)+1):
            V=V+1
            H=H+1/V**K1
        V=V+1
    R=1/V**2
    P=R*C[K][6-1]
    for I in range(5,1-1,-1):
        P=R*(C[K][I-1]+P)
    H=SGN[K]*(FCT[K+1]*H+(V*(FCT[K-1+1]+P)+HF*FCT[K+1])/V**K1)
    if K == 0:
        H=H+np.log(V)
    if X < 0:
        V=np.pi*U
        X=np.real(V)
        Y=np.imag(V)
        A=np.sin(X)
        B=np.cos(X)
        T=np.tanh(Y)
        P=complex(B,-A*T)/complex(A,B*T)
        if K == 0:
            H=H+1/U+np.pi*P
        elif K == 1:
            H=-H+1/U**2+C1*(P**2+1)
        elif K == 2:
            H=H+2/U**3+C2*P*(P**2+1)
        elif K == 3:
            R=P**2
            H=-H+6/U**4+C3*((3*R+4)*R+1)
        elif K == 4:
            R=P**2
            H=H+24/U**5+C4*P*((3*R+5)*R+2)
    return H
    # fmt: on


@nb.njit(cache=True)
def recursive_harmonic_sum(base_value, n, iterations, weight):
    """
    Compute the harmonic sum :math:`S_{w}(N+k)` stating from the value
    :math:`S_{w}(N)` via the recurrence relations.

    Parameters
    ----------
        base_value: complex
            starting value :math:`S_{w}(N)`
        n: complex
            starting point
        iterations: int
            number of iterations
        weight: int
            harmonic sum weight

    Returns
    -------
        sni : complex
            :math:`S_{w}(N+k)`
    """
    fact = 0.0
    for i in range(1, iterations + 1):
        fact += 1.0 / (n + i) ** weight
    return base_value + fact


@nb.njit(cache=True)
def symmetry_factor(N, is_singlet=None):
    """
    Compute the analytical continuation of :math:`(-1)^N`

    Parameters
    ----------
        N: complex
            Mellin moment
        is_singlet: bool, None
            True for singlet like quantities
            False for non-singlet like quantities
            None for generic complex N value

    Returns
    -------
        eta: complex
            1 for singlet like quantities,
            -1 for non-singlet like quantities,
            :math:`(-1)^N` elsewise
    """
    if is_singlet is None:
        return (-1) ** N
    if is_singlet:
        return 1
    return -1



# -*- coding: utf-8 -*-
"""
LHAPDF interface
"""
import numpy as np

from eko import basis_rotation as br


def compute_LHAPDF_data(operators, pdf, skip_pdfs, rotate_to_evolution_basis=False):
    """
    Run LHAPDF to compute operators.

    Parameters
    ----------
        operators : dict
            operators card
        pdf : lhapdf_type
            pdf
        skip_pdfs : list
            list of pdfs (pid or name) to skip
        rotate_to_evolution_basis: bool
            rotate to evolution basis

    Returns
    -------
        ref : dict
            output containing: target_xgrid, values
    """

    target_xgrid = operators["interpolation_xgrid"]

    out_tabs = {}
    for q2 in operators["Q2grid"]:

        tab = {}
        for pid in br.flavor_basis_pids:

            if pid in skip_pdfs:
                continue

            # collect lhapdf
            me = []
            for x in target_xgrid:
                xf = pdf.xfxQ2(pid, x, q2)
                me.append(xf)
            tab[pid] = np.array(me)

        # rotate if needed
        if rotate_to_evolution_basis:
            pdfs = np.array(
                [
                    tab[pid] if pid in tab else np.zeros(len(target_xgrid))
                    for pid in br.flavor_basis_pids
                ]
            )
            evol_pdf = br.rotate_flavor_to_evolution @ pdfs
            tab = dict(zip(br.evol_basis, evol_pdf))

        out_tabs[q2] = tab

    ref = {
        "target_xgrid": target_xgrid,
        "values": out_tabs,
    }

    return ref



# -*- coding: utf-8 -*-
import pathlib

import eko
from eko import basis_rotation as br
from ekomark import apply

from . import gen_info, genpdf


def evolve_pdfs(
    initial_PDF_list,
    theory_card,
    operators_card,
    path=None,
    store_path=None,
    targetgrid=None,
    install=False,
    name="Evolved_PDF",
    info_update=None,
):
    """
    This function evolves an initial_PDF using a theory card and an operator card
    and dump the evolved PDF in lhapdf format

    Parameters
    ----------
        initial_PDF_list : list(lhapdf object)
            list of PDF members to be evolved

        theory_card : dict
            theory card

        operators_card : dict
            operators card

        path : str
            path to cached eko output (if "None" it will be recomputed)

        store_path : str
            path where the eko is stored (if "None" will not be saved)

        targetgrid : list(float)
            target x-grid (if different from input x-grid)

        install : bool
            set whether to install evolved PDF to lhapdf directory

        name : str
            set name of evolved PDF

        info_update : dict
            dict of info to add or update to default info file

    """
    eko_output = None
    if path is not None:
        my_path = pathlib.Path(path)
        if my_path.is_dir():
            ops_id = f"o{operators_card['hash'][:6]}_t{theory_card['hash'][:6]}.tar"
            ops_id_path = pathlib.Path(ops_id)
            outpath = my_path / ops_id_path.relative_to(ops_id_path.anchor)
            eko_output = eko.output.Output.load_tar(outpath)
        else:
            eko_output = eko.output.Output.load_tar(my_path)
    else:
        eko_output = eko.run_dglap(theory_card, operators_card)
        if store_path is not None:
            eko_output.dump_tar(store_path)

    evolved_PDF_list = []
    for initial_PDF in initial_PDF_list:
        evolved_PDF_list.append(apply.apply_pdf(eko_output, initial_PDF, targetgrid))

    if targetgrid is None:
        targetgrid = operators_card["interpolation_xgrid"]
    if info_update is None:
        info_update = {}
    info_update["XMin"] = targetgrid[0]
    info_update["XMax"] = targetgrid[-1]
    info = gen_info.create_info_file(
        theory_card,
        operators_card,
        len(evolved_PDF_list),
        info_update=info_update,
    )
    all_member_blocks = []
    for evolved_PDF in evolved_PDF_list:
        all_blocks = []
        block = genpdf.generate_block(
            lambda pid, x, Q2, evolved_PDF=evolved_PDF: targetgrid[targetgrid.index(x)]
            * evolved_PDF[Q2]["pdfs"][pid][targetgrid.index(x)],
            xgrid=targetgrid,
            Q2grid=operators_card["Q2grid"],
            pids=br.flavor_basis_pids,
        )
        # all_blocks will be useful in case there will be necessity to dump many blocks
        # for a single member
        all_blocks.append(block)
        all_member_blocks.append(all_blocks)

    genpdf.export.dump_set(name, info, all_member_blocks)

    if install:
        genpdf.install_pdf(name)


def gen_out(theory_card, op_card, path=None):
    """
    Generates EKO output from theory and operators cards and, if requested,
    dumps it in tar format

    Parameters
    ----------
        theory_card : dict
            theory card
        op_card : dict
            operators card
        path : str
            path of dumped output (if "None" output is not dumped)

    Returns
    -------
        : eko.output.Output
            eko output
    """
    eko_output = eko.run_dglap(theory_card, op_card)
    if path is not None:
        ops_id = f"o{op_card['hash'][:6]}_t{theory_card['hash'][:6]}"
        path = f"{path}/{ops_id}.tar"
        eko_output.dump_tar(path)
    return eko_output



# -*- coding: utf-8 -*-

from . import output, runner, version

__version__ = version.__version__


def run_dglap(theory_card, operators_card):
    r"""
    This function takes a DGLAP theory configuration dictionary
    and performs the solution of the DGLAP equations.

    The EKO :math:`\mathbf E_{k,j}(a_s^1\leftarrow a_s^0)` is determined in order
    to fullfill the following evolution

    .. math::
        \mathbf f(x_k,a_s^1) = \mathbf E_{k,j}(a_s^1\leftarrow a_s^0) \mathbf f(x_j,a_s^0)

    Parameters
    ----------
        setup : dict
            input card - see :doc:`/code/IO`

    Returns
    -------
        output : dict
            output dictionary - see :doc:`/code/IO`
    """
    r = runner.Runner(theory_card, operators_card)
    output = r.get_output()
    return output



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['eko',
 'eko.anomalous_dimensions',
 'eko.anomalous_dimensions.as4',
 'eko.evolution_operator',
 'eko.harmonics',
 'eko.harmonics.f_functions',
 'eko.kernels',
 'eko.matching_conditions',
 'eko.matching_conditions.as3',
 'eko.scale_variations',
 'ekobox',
 'ekobox.genpdf',
 'ekomark',
 'ekomark.benchmark',
 'ekomark.benchmark.external',
 'ekomark.data',
 'ekomark.navigator']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'lz4>=3.1.10,<4.0.0',
 'numba>=0.55.0,<0.56.0',
 'numpy>=1.22,<2.0',
 'scipy>=1.7.3,<2.0.0']

extras_require = \
{'box': ['banana-hep>=0.6.6,<0.7.0',
         'sqlalchemy>=1.4.21,<2.0.0',
         'pandas>=1.3.0,<2.0.0',
         'matplotlib>=3.5.1,<4.0.0'],
 'docs': ['Sphinx>=4.3.2,<5.0.0',
          'sphinx-rtd-theme>=1.0.0,<2.0.0',
          'sphinxcontrib-bibtex>=2.4.1,<3.0.0',
          'nbsphinx>=0.8.8,<0.9.0'],
 'mark': ['banana-hep>=0.6.6,<0.7.0',
          'sqlalchemy>=1.4.21,<2.0.0',
          'pandas>=1.3.0,<2.0.0',
          'matplotlib>=3.5.1,<4.0.0']}

entry_points = \
{'console_scripts': ['ekonav = ekomark.navigator:launch_navigator',
                     'genpdf = ekobox.genpdf.cli:cli']}

setup_kwargs = {
    'name': 'eko',
    'version': '0.10.2',
    'description': 'Evolution Kernel Operators',
    'long_description': '<p align="center">\n  <a href="https://eko.readthedocs.io/"><img alt="EKO" src="https://raw.githubusercontent.com/N3PDF/eko/master/doc/source/img/Logo.png" width=300></a>\n</p>\n<p align="center">\n  <a href="https://github.com/N3PDF/eko/actions/workflows/unittests.yml"><img alt="Tests" src="https://github.com/N3PDF/eko/actions/workflows/unittests.yml/badge.svg" /></a>\n  <a href="https://eko.readthedocs.io/en/latest/?badge=latest"><img alt="Docs" src="https://readthedocs.org/projects/eko/badge/?version=latest"></a>\n  <a href="https://codecov.io/gh/N3PDF/eko"><img src="https://codecov.io/gh/N3PDF/eko/branch/master/graph/badge.svg" /></a>\n  <a href="https://www.codefactor.io/repository/github/n3pdf/eko"><img src="https://www.codefactor.io/repository/github/n3pdf/eko/badge" alt="CodeFactor" /></a>\n</p>\n\nEKO is a Python module to solve the DGLAP equations in N-space in terms of Evolution Kernel Operators in x-space.\n\n## Installation\nEKO is available via\n- PyPI: <a href="https://pypi.org/project/eko/"><img alt="PyPI" src="https://img.shields.io/pypi/v/eko"/></a>\n```bash\npip install eko\n```\n- conda-forge: [![Conda Version](https://img.shields.io/conda/vn/conda-forge/eko.svg)](https://anaconda.org/conda-forge/eko)\n```bash\nconda install eko\n```\n\n### Development\n\nIf you want to install from source you can run\n```bash\ngit clone git@github.com:N3PDF/eko.git\ncd eko\npoetry install\n```\n\nTo setup `poetry`, and other tools, see [Contribution\nGuidelines](https://github.com/N3PDF/eko/blob/master/.github/CONTRIBUTING.md).\n\n## Documentation\n- The documentation is available here: <a href="https://eko.readthedocs.io/en/latest/?badge=latest"><img alt="Docs" src="https://readthedocs.org/projects/eko/badge/?version=latest"></a>\n- To build the documentation from source install [graphviz](https://www.graphviz.org/) and run in addition to the installation commands\n```bash\npoe docs\n```\n\n## Tests and benchmarks\n- To run unit test you can do\n```bash\npoe tests\n```\n\n- Benchmarks of specific part of the code, such as the strong coupling or msbar masses running, are available doing\n```bash\npoe bench\n```\n\n- The complete list of benchmarks with external codes is available through `ekomark`: [documentation](https://eko.readthedocs.io/en/latest/development/Benchmarks.html)\n\n## Citation policy\nWhen using our code please cite\n- our DOI: <a href="https://doi.org/10.5281/zenodo.3874237"><img src="https://zenodo.org/badge/DOI/10.5281/zenodo.3874237.svg" alt="DOI"/></a>\n- our paper: [![arXiv](https://img.shields.io/badge/arXiv-2202.02338-b31b1b?labelColor=222222)](https://arxiv.org/abs/2202.02338)\n\n## Contributing\n- Your feedback is welcome! If you want to report a (possible) bug or want to ask for a new feature, please raise an issue: <a href="https://github.com/N3PDF/eko/issues"><img alt="GitHub issues" src="https://img.shields.io/github/issues/N3PDF/eko"/></a>\n- Please follow our [Code of Conduct](https://github.com/N3PDF/eko/blob/master/.github/CODE_OF_CONDUCT.md) and read the\n  [Contribution Guidelines](https://github.com/N3PDF/eko/blob/master/.github/CONTRIBUTING.md)\n',
    'author': 'A. Barontini',
    'author_email': 'andrea.barontini@mi.infn.it',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/N3PDF/eko',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""This files sets the physical constants."""

import numba as nb

NC = 3
"""The number of colors."""

TR = float(1.0 / 2.0)
"""The normalization of fundamental generators.

Defaults to :math:`T_R = 1/2`.
"""

CA = float(NC)
"""Second Casimir constant in the adjoint representation.

Defaults to :math:`N_C = 3`.
"""

CF = float((NC * NC - 1.0) / (2.0 * NC))
r"""Second Casimir constant in the fundamental representation.

Defaults to :math:`\frac{N_C^2-1}{2N_C} = 4/3`.
"""

eu2 = 4.0 / 9
"""Up quarks charge squared."""

ed2 = 1.0 / 9
"""Down quarks charge squared."""


def update_colors(nc):
    """Updates the number of colors to :math:`NC = nc`.

    The Casimirs for a generic value of :math:`NC` are consistenly updated as
    well.

    Parameters
    ----------
    nc : int
      Number of colors

    """
    global NC, CA, CF  # pylint: disable=global-statement

    NC = int(nc)
    CA = float(NC)
    CF = float((NC * NC - 1.0) / (2.0 * NC))


@nb.njit(cache=True)
def uplike_flavors(nf):
    """Computes the number of up flavors

    Parameters
    ----------
        nf : int
            Number of active flavors

    Returns
    -------
        nu : int

    """
    if nf not in range(2, 6 + 1):
        raise NotImplementedError("Selected nf is not implemented")
    nu = nf // 2
    return nu



# -*- coding: utf-8 -*-
"""
This file contains the O(aem1) Altarelli-Parisi splitting kernels.
"""

import numba as nb

from .. import constants
from . import as1


@nb.njit(cache=True)
def gamma_phq(N):
    """
    Computes the leading-order photon-quark anomalous dimension

    Implements Eq. (2.5) of :cite:`Carrazza:2015dea`.

    Parameters
    ----------
      N : complex
        Mellin moment

    Returns
    -------
      gamma_phq : complex
        Leading-order photon-quark anomalous dimension :math:`\\gamma_{\\gamma q}^{(0)}(N)`
    """

    return as1.gamma_gq(N) / constants.CF


@nb.njit(cache=True)
def gamma_qph(N, nf):
    """
    Computes the leading-order quark-photon anomalous dimension

    Implements Eq. (2.5) of :cite:`Carrazza:2015dea`.
    But adding the :math:`N_C` and the :math:`2n_f` factors from :math:`\\theta` inside the
    definition of :math:`\\gamma_{q \\gamma}^{(0)}(N)`.

    Parameters
    ----------
      N : complex
        Mellin moment
      nf : int
        Number of active flavors

    Returns
    -------
      gamma_qph : complex
        Leading-order quark-photon anomalous dimension :math:`\\gamma_{q \\gamma}^{(0)}(N)`
    """
    return as1.gamma_qg(N, nf) / constants.TR * constants.NC


@nb.njit(cache=True)
def gamma_phph(nf):
    """
    Computes the leading-order photon-photon anomalous dimension

    Implements Eq. (2.5) of :cite:`Carrazza:2015dea`.

    Parameters
    ----------
      nf : int
        Number of active flavors

    Returns
    -------
      gamma_phph : complex
        Leading-order phton-photon anomalous dimension :math:`\\gamma_{\\gamma \\gamma}^{(0)}(N)`
    """

    return 2 / 3 * constants.NC * 2 * nf


@nb.njit(cache=True)
def gamma_ns(N, s1):
    """
    Computes the leading-order non-singlet QED anomalous dimension.

    Implements Eq. (2.5) of :cite:`Carrazza:2015dea`.

    Parameters
    ----------
      N : complex
        Mellin moment
      s1 : complex
        S1(N)

    Returns
    -------
      gamma_ns : complex
        Leading-order non-singlet QED anomalous dimension :math:`\\gamma_{ns}^{(0)}(N)`
    """
    return as1.gamma_ns(N, s1) / constants.CF



# -*- coding: utf-8 -*-
"""
Additional package to benchmark eko.
"""
from eko import basis_rotation as br


def pdfname(pid_or_name):
    """Return pdf name"""
    if isinstance(pid_or_name, int):
        return br.flavor_basis_names[br.flavor_basis_pids.index(pid_or_name)]
    return pid_or_name



# -*- coding: utf-8 -*-
"""
This module defines the |OME| for the non-trivial matching conditions in the
|VFNS| evolution.
"""

import functools
import logging

import numba as nb
import numpy as np

from .. import basis_rotation as br
from .. import harmonics
from ..evolution_operator import Operator, QuadKerBase
from . import as1, as2, as3

logger = logging.getLogger(__name__)


@nb.njit(cache=True)
def A_singlet(matching_order, n, sx, nf, L, is_msbar, sx_ns=None):
    r"""
    Computes the tower of the singlet |OME|.

    Parameters
    ----------
        matching_order : tuple(int,int)
            perturbative matching_order
        n : complex
            Mellin variable
        sx : list
            singlet like harmonic sums cache
        nf: int
            number of active flavor below threshold
        L : float
            :math:`log(q^2/m_h^2)`
        is_msbar: bool
            add the |MSbar| contribution
        sx_ns : list
            non-singlet like harmonic sums cache

    Returns
    -------
        A_singlet : numpy.ndarray
            singlet |OME|

    See Also
    --------
        eko.matching_conditions.nlo.A_singlet_1 : :math:`A^{S,(1)}(N)`
        eko.matching_conditions.nlo.A_hh_1 : :math:`A_{HH}^{(1)}(N)`
        eko.matching_conditions.nlo.A_gh_1 : :math:`A_{gH}^{(1)}(N)`
        eko.matching_conditions.nnlo.A_singlet_2 : :math:`A_{S,(2)}(N)`
    """
    A_s = np.zeros((matching_order[0], 3, 3), np.complex_)
    if matching_order[0] >= 1:
        A_s[0] = as1.A_singlet(n, sx, L)
    if matching_order[0] >= 2:
        A_s[1] = as2.A_singlet(n, sx, L, is_msbar)
    if matching_order[0] >= 3:
        A_s[2] = as3.A_singlet(n, sx, sx_ns, nf, L)
    return A_s


@nb.njit(cache=True)
def A_non_singlet(matching_order, n, sx, nf, L):
    r"""
    Computes the tower of the non-singlet |OME|

    Parameters
    ----------
        matching_order : tuple(int,int)
            perturbative matching_order
        n : complex
            Mellin variable
        sx : list
            harmonic sums cache
        nf: int
            number of active flavor below threshold
        L : float
            :math:`log(q^2/m_h^2)`

    Returns
    -------
        A_non_singlet : numpy.ndarray
            non-singlet |OME|

    See Also
    --------
        eko.matching_conditions.nlo.A_hh_1 : :math:`A_{HH}^{(1)}(N)`
        eko.matching_conditions.nnlo.A_ns_2 : :math:`A_{qq,H}^{NS,(2)}`
    """
    A_ns = np.zeros((matching_order[0], 2, 2), np.complex_)
    if matching_order[0] >= 1:
        A_ns[0] = as1.A_ns(n, sx, L)
    if matching_order[0] >= 2:
        A_ns[1] = as2.A_ns(n, sx, L)
    if matching_order[0] >= 3:
        A_ns[2] = as3.A_ns(n, sx, nf, L)
    return A_ns


@nb.njit(cache=True)
def build_ome(A, matching_order, a_s, backward_method):
    r"""
    Construct the matching expansion in :math:`a_s` with the appropriate method.

    Parameters
    ----------
        A : numpy.ndarray
            list of |OME|
        matching_order : tuple(int,int)
            perturbation matching order
        a_s : float
            strong coupling, needed only for the exact inverse
        backward_method : ["exact", "expanded" or ""]
            empty or method for inverting the matching condition (exact or expanded)

    Returns
    -------
        ome : numpy.ndarray
            matching operator matrix
    """
    # to get the inverse one can use this FORM snippet
    # Symbol a;
    # NTensor c,d,e;
    # Local x=-(a*c+a**2* d + a**3 * e);
    # Local bi = 1+x+x**2+x**3;
    # Print;
    # .end
    ome = np.eye(len(A[0]), dtype=np.complex_)
    A = np.ascontiguousarray(A)
    if backward_method == "expanded":
        # expended inverse
        if matching_order[0] >= 1:
            ome -= a_s * A[0]
        if matching_order[0] >= 2:
            ome += a_s**2 * (-A[1] + A[0] @ A[0])
        if matching_order[0] >= 3:
            ome += a_s**3 * (-A[2] + A[0] @ A[1] + A[1] @ A[0] - A[0] @ A[0] @ A[0])
    else:
        # forward or exact inverse
        if matching_order[0] >= 1:
            ome += a_s * A[0]
        if matching_order[0] >= 2:
            ome += a_s**2 * A[1]
        if matching_order[0] >= 3:
            ome += a_s**3 * A[2]
        # need inverse exact ?  so add the missing pieces
        if backward_method == "exact":
            ome = np.linalg.inv(ome)
    return ome


@nb.njit(cache=True)
def quad_ker(
    u, order, mode0, mode1, is_log, logx, areas, a_s, nf, L, backward_method, is_msbar
):
    """
    Raw kernel inside quad

    Parameters
    ----------
        u : float
            quad argument
        order : tuple(int,int)
            perturbation matching order
        mode0 : int
            pid for first element in the singlet sector
        mode1 : int
            pid for second element in the singlet sector
        is_log : boolean
            logarithmic interpolation
        logx : float
            Mellin inversion point
        areas : tuple
            basis function configuration
        a_s : float
            strong coupling, needed only for the exact inverse
        nf: int
            number of active flavor below threshold
        L : float
            :math:`log(q^2/m_h^2)`
        backward_method : ["exact", "expanded" or ""]
            empty or method for inverting the matching condition (exact or expanded)
        is_msbar: bool
            add the |MSbar| contribution
    Returns
    -------
        ker : float
            evaluated integration kernel
    """
    ker_base = QuadKerBase(u, is_log, logx, mode0)
    integrand = ker_base.integrand(areas)
    if integrand == 0.0:
        return 0.0

    max_weight_dict = {1: 2, 2: 3, 3: 5}
    sx = harmonics.compute_cache(
        ker_base.n, max_weight_dict[order[0]], ker_base.is_singlet
    )
    sx_ns = sx.copy()
    if order[0] == 3 and (
        (backward_method != "" and ker_base.is_singlet)
        or (mode0 == 100 and mode1 == 100)
    ):
        # At N3LO for A_qq singlet or backward you need to compute
        # both the singlet and non-singlet like harmonics
        # avoiding recomputing all of them ...
        smx_ns = harmonics.smx(ker_base.n, np.array([s[0] for s in sx]), False)
        for w, sm in enumerate(smx_ns):
            sx_ns[w][-1] = sm
        sx_ns[2][2] = harmonics.S2m1(ker_base.n, sx[0][1], smx_ns[0], smx_ns[1], False)
        sx_ns[2][3] = harmonics.Sm21(ker_base.n, sx[0][0], smx_ns[0], False)
        sx_ns[3][5] = harmonics.Sm31(ker_base.n, sx[0][0], smx_ns[0], smx_ns[1], False)
        sx_ns[3][4] = harmonics.Sm211(ker_base.n, sx[0][0], sx[0][1], smx_ns[0], False)
        sx_ns[3][3] = harmonics.Sm22(
            ker_base.n, sx[0][0], sx[0][1], smx_ns[1], sx_ns[3][5], False
        )

    # compute the ome
    if ker_base.is_singlet:
        indices = {21: 0, 100: 1, 90: 2}
        A = A_singlet(order, ker_base.n, sx, nf, L, is_msbar, sx_ns)
    else:
        indices = {200: 0, 91: 1}
        A = A_non_singlet(order, ker_base.n, sx, nf, L)

    # build the expansion in alpha_s depending on the strategy
    ker = build_ome(A, order, a_s, backward_method)

    # select the needed matrix element
    ker = ker[indices[mode0], indices[mode1]]

    # recombine everything
    return np.real(ker * integrand)


class OperatorMatrixElement(Operator):
    """Internal representation of a single |OME|.

    The actual matrices are computed upon calling :meth:`compute`.

    Parameters
    ----------
    config : dict
        configuration
    managers : dict
        managers
    nf: int
        number of active flavor below threshold
    q2: float
        matching scale
    is_backward: bool
        True for backward evolution
    L: float
        log of K threshold squared
    is_msbar: bool
        add the |MSbar| contribution
    """

    log_label = "Matching"
    # complete list of possible matching operators labels
    full_labels = [
        *br.singlet_labels,
        (br.matching_hplus_pid, 21),
        (br.matching_hplus_pid, 100),
        (21, br.matching_hplus_pid),
        (100, br.matching_hplus_pid),
        (br.matching_hplus_pid, br.matching_hplus_pid),
        (200, 200),
        (200, br.matching_hminus_pid),
        (br.matching_hminus_pid, 200),
        (br.matching_hminus_pid, br.matching_hminus_pid),
    ]

    def __init__(self, config, managers, nf, q2, is_backward, L, is_msbar):
        super().__init__(config, managers, nf, q2, None)
        self.backward_method = config["backward_inversion"] if is_backward else ""
        if is_backward:
            self.is_intrinsic = True
        else:
            self.is_intrinsic = bool(len(config["intrinsic_range"]) != 0)
        self.L = L
        self.is_msbar = is_msbar
        # Note for the moment only QCD matching is implemented
        self.order = (self.order[0] - 1, self.order[1])

    @property
    def labels(self):
        """Computes the necessary sector labels to compute.

        Returns
        -------
        list(str)
            sector labels
        """

        labels = []
        # non-singlet labels
        if self.config["debug_skip_non_singlet"]:
            logger.warning("%s: skipping non-singlet sector", self.log_label)
        else:
            labels.append((200, 200))
            if self.is_intrinsic or self.backward_method != "":
                # intrinsic labels, which are not zero at NLO
                labels.append((br.matching_hminus_pid, br.matching_hminus_pid))
                # These contributions are always 0 for the moment
                # labels.extend([(200, br.matching_hminus_pid), (br.matching_hminus_pid, 200)])
        # same for singlet
        if self.config["debug_skip_singlet"]:
            logger.warning("%s: skipping singlet sector", self.log_label)
        else:
            labels.extend(
                [
                    *br.singlet_labels,
                    (br.matching_hplus_pid, 21),
                    (br.matching_hplus_pid, 100),
                ]
            )
            if self.is_intrinsic or self.backward_method != "":
                labels.extend(
                    [
                        (21, br.matching_hplus_pid),
                        (100, br.matching_hplus_pid),
                        (br.matching_hplus_pid, br.matching_hplus_pid),
                    ]
                )
        return labels

    def quad_ker(self, label, logx, areas):
        """Partially initialized integrand function.

        Parameters
        ----------
        label: tuple
            operator element pids
        logx: float
            Mellin inversion point
        areas : tuple
            basis function configuration

        Returns
        -------
        functools.partial
            partially initialized integration kernel
        """
        return functools.partial(
            quad_ker,
            order=self.order,
            mode0=label[0],
            mode1=label[1],
            is_log=self.int_disp.log,
            logx=logx,
            areas=areas,
            a_s=self.a_s,
            nf=self.nf,
            L=self.L,
            backward_method=self.backward_method,
            is_msbar=self.is_msbar,
        )

    @property
    def a_s(self):
        """Returns the computed values for :math:`a_s`.

        Note that here you need to use :math:`a_s^{n_f+1}`
        """
        sc = self.managers["strong_coupling"]
        return sc.a_s(self.mur2_shift(self.q2_from), self.q2_from, nf_to=self.nf + 1)

    def compute(self):
        """Compute the actual operators (i.e. run the integrations)"""
        self.initialize_op_members()

        # At LO you don't need anything else
        if self.order[0] == 0:
            logger.info("%s: no need to compute matching at LO", self.log_label)
            return

        self.integrate()



# -*- coding: utf-8 -*-
"""
    This file contains the main application class of eko
"""
import copy
import logging

import numpy as np

from . import basis_rotation as br
from . import compatibility, interpolation, msbar_masses
from .couplings import Couplings
from .evolution_operator.grid import OperatorGrid
from .output import Output
from .thresholds import ThresholdsAtlas

logger = logging.getLogger(__name__)


class Runner:
    """
    Represents a single input configuration.

    For details about the configuration, see :doc:`here </code/IO>`

    Parameters
    ----------
        setup : dict
            input configurations
    """

    banner = r"""
oooooooooooo oooo    oooo  \\ .oooooo.
`888'     `8 `888   .8P'  //////    `Y8b
 888          888  d8'   \\o\/////    888
 888oooo8     88888     \\\\/8/////   888
 888    "     888`88b.      888 ///   888
 888       o  888  `88b.    `88b //  d88'
o888ooooood8 o888o  o888o     `Y8bood8P'
"""

    def __init__(self, theory_card, operators_card):
        self.out = Output()

        new_theory, new_operators = compatibility.update(theory_card, operators_card)

        # Store inputs
        self._theory = new_theory

        # setup basis grid
        bfd = interpolation.InterpolatorDispatcher.from_dict(new_operators)
        self.out.update(bfd.to_dict())
        # setup the Threshold path, compute masses if necessary
        masses = None
        if new_theory["HQ"] == "MSBAR":
            masses = msbar_masses.compute(new_theory)
        tc = ThresholdsAtlas.from_dict(new_theory, masses=masses)

        self.out["q2_ref"] = float(tc.q2_ref)
        # strong coupling
        sc = Couplings.from_dict(new_theory, masses=masses)
        # setup operator grid
        self.op_grid = OperatorGrid.from_dict(
            new_theory,
            new_operators,
            tc,
            sc,
            bfd,
        )
        self.out["inputgrid"] = bfd.xgrid_raw
        self.out["targetgrid"] = bfd.xgrid_raw
        self.post_process = dict(
            inputgrid=new_operators.get("inputgrid", bfd.xgrid_raw),
            targetgrid=new_operators.get("targetgrid", bfd.xgrid_raw),
            inputbasis=new_operators.get("inputbasis"),
            targetbasis=new_operators.get("targetbasis"),
        )

    def get_output(self):
        """
        Collects all data for output (to run the evolution)

        Returns
        -------
            ret : eko.output.Output
                output instance
        """
        # add all operators
        Q2grid = {}
        self.out["inputpids"] = br.flavor_basis_pids
        self.out["targetpids"] = br.flavor_basis_pids
        self.out["inputgrid"] = self.out["interpolation_xgrid"]
        self.out["targetgrid"] = self.out["interpolation_xgrid"]
        for final_scale, op in self.op_grid.compute().items():
            Q2grid[float(final_scale)] = op
        self.out["Q2grid"] = Q2grid
        # reshape xgrid
        inputgrid = (
            self.post_process["inputgrid"]
            if self.post_process["inputgrid"] is not self.out["interpolation_xgrid"]
            else None
        )
        targetgrid = (
            self.post_process["targetgrid"]
            if self.post_process["targetgrid"] is not self.out["interpolation_xgrid"]
            else None
        )
        if inputgrid is not None or targetgrid is not None:
            self.out.xgrid_reshape(targetgrid=targetgrid, inputgrid=inputgrid)

        # reshape flavors
        inputbasis = self.post_process["inputbasis"]
        if inputbasis is not None:
            inputbasis = np.array(inputbasis)
        targetbasis = self.post_process["targetbasis"]
        if targetbasis is not None:
            targetbasis = np.array(targetbasis)
        if inputbasis is not None or targetbasis is not None:
            self.out.flavor_reshape(targetbasis=targetbasis, inputbasis=inputbasis)
        return copy.deepcopy(self.out)



# -*- coding: utf-8 -*-
"""This module contain the possible scale variations schemes."""
import enum

from . import expanded, exponentiated


class Modes(enum.IntEnum):
    """Scale Variation modes"""

    unvaried = enum.auto()
    exponentiated = enum.auto()
    expanded = enum.auto()


def sv_mode(s):
    """Returns the scale variation mode.

    Parameters
    ----------
    s : str
        string representation

    Returns
    -------
    enum.IntEnum
        enum representation
    """
    if s is not None:
        return Modes[s]
    return Modes.unvaried


class ModeMixin:
    """Mixin to cast scale variation mode."""

    @property
    def sv_mode(self):
        """Returns the scale variation mode"""
        return sv_mode(self.config["ModSV"])



# -*- coding: utf-8 -*-
import copy
import pathlib
import shutil

import numpy as np
from banana import toy

from eko import basis_rotation as br

from . import export, flavors, load


def take_data(parent_pdf_set=None, members=False):
    """
    Auxiliary function for generate_pdf. It provides the info, the heads
    of the member files and the blocks to be generated to generate_pdf.

    Parameters
    ----------
        parent_pdf_set :
            the PDF set to be used as parent
        members : bool
            if true every member of the parent are loaded
    """
    xgrid = np.geomspace(1e-9, 1, 240)
    Q2grid = np.geomspace(1.3, 1e5, 35)
    # collect blocks
    all_blocks = []
    info = None
    heads = []
    if parent_pdf_set is None:
        parent_pdf_set = {
            pid: lambda x, _Q2: x * (1 - x) for pid in br.flavor_basis_pids
        }
    if isinstance(parent_pdf_set, str):
        if parent_pdf_set in ["toylh", "toy"]:
            info = copy.deepcopy(load.Toy_info)
            toylh = toy.mkPDF("", 0)
            all_blocks.append(
                [generate_block(toylh.xfxQ2, xgrid, Q2grid, br.flavor_basis_pids)]
            )
        else:
            info = load.load_info_from_file(parent_pdf_set)
            # iterate on members
            for m in range(int(info["NumMembers"])):
                dat = load.load_blocks_from_file(parent_pdf_set, m)
                heads.append(dat[0])
                all_blocks.append(dat[1])
                if not members:
                    break
    elif isinstance(parent_pdf_set, dict):
        info = copy.deepcopy(load.template_info)
        all_blocks.append(
            [
                generate_block(
                    lambda pid, x, Q2: 0.0
                    if pid not in parent_pdf_set
                    else parent_pdf_set[pid](x, Q2),
                    xgrid,
                    Q2grid,
                    br.flavor_basis_pids,
                )
            ]
        )
    else:
        raise ValueError("Unknown parent pdf type")
    return heads, info, all_blocks


def generate_pdf(
    name, labels, parent_pdf_set=None, members=False, info_update=None, install=False
):
    """
    Generate a new PDF from a parent PDF with a set of flavors.

    If parent_pdf_set is the name of an available PDF set,
    it will be used as parent. In order to use the toy PDF
    as parent, it is enough to set parent_pdf_set = "toy" or "toylh".
    If parent_pdf_set is not specified, a debug PDF constructed as
    x * (1-x) for every flavor will be usedas parent.
    It is also possible to provide custom functions for each flavor
    in the form of a dictionary: {pid: f(x,Q2)}.

    In labels it is possible to pass a list of PIDs or evolution basis
    combinations to keep in the generated PDF. In order to project
    on custom combinations of PIDs, it is also possible to pass a list
    containing the desired factors for each flavor.

    The default behaviour is to generate only one member for a PDF set
    (the zero member) but it can be changed setting to True the member flag.

    The info_update argument is a dictionary and provide to the user a way
    to change the info file of the generated PDF set. If a key of info_update
    matches with one key of the standard info file, the information
    are updated, otherwise they are simply added.

    Turning True the value of the install flag, it is possible to autmatically
    install the generated PDF to the lhapdf directory. By default install is False.

    Parameters
    ----------
        name : str
            target name
        labels :
            list of flavors
        parent_pdf_set :
            parent PDF
        all : bool
            iterate on members
        install : bool
            install on LHAPDF path

    Examples
    --------
        To generate a PDF with a fixed function `f(x,Q2)` for some flavors
        you can use the following snippet:

        >>> # f = lambda x,Q2 ... put the desired function here
        >>> # mask = [list of active PIDs]
        >>> generate_pdf(name, labels, parent_pdf_set={pid: f for pid in mask})

        The |API| also provides the possibility to extract arbitrary flavor combinations:
        using the debug PDF settings we can construct a "anti-QED-singlet" combination that
        is usefull in debugging DIS codes since it does not couple in |LO|, but only
        through the pure-singlet contributions (starting at |NNLO|)

        >>> from eko import basis_rotation as br
        >>> from ekobox.tools import genpdf
        >>> import numpy as np
        >>> anti_qed_singlet = np.zeros_like(br.flavor_basis_pids, dtype=np.float_)
        >>> anti_qed_singlet[br.flavor_basis_pids.index(1)] = -4
        >>> anti_qed_singlet[br.flavor_basis_pids.index(-1)] = -4
        >>> anti_qed_singlet[br.flavor_basis_pids.index(2)] = 1
        >>> anti_qed_singlet[br.flavor_basis_pids.index(-2)] = 1
        >>> genpdf.generate_pdf("anti_qed_singlet", [anti_qed_singlet])
    """
    pathlib.Path(name).mkdir(exist_ok=True)
    # Checking label basis
    is_evol = False
    flavor_combinations = labels
    if flavors.is_evolution_labels(labels):
        is_evol = True
        flavor_combinations = flavors.evol_to_flavor(labels)
    elif flavors.is_pid_labels(labels):
        labels = np.array(labels, dtype=np.int_)
        flavor_combinations = flavors.pid_to_flavor(labels)

    # labels = verify_labels(args.labels)
    heads, info, all_blocks = take_data(parent_pdf_set=parent_pdf_set, members=members)

    # filter the PDF
    new_all_blocks = []
    for b in all_blocks:
        new_all_blocks.append(flavors.project(b, flavor_combinations))

    # changing info file according to user choice
    if info_update is not None:
        if isinstance(info_update, dict):
            info.update(info_update)
        else:
            raise TypeError("Info to update are not in a dictionary format")
    # write
    info["Flavors"] = [int(pid) for pid in br.flavor_basis_pids]
    info["NumFlavors"] = len(br.flavor_basis_pids)
    if is_evol:
        info["ForcePositive"] = 0
    info["NumMembers"] = len(new_all_blocks)
    # exporting
    export.dump_set(name, info, new_all_blocks, pdf_type_list=heads)

    # install
    if install:
        install_pdf(name)


def install_pdf(name):
    """
    Install set into LHAPDF.

    The set to be installed has to be in the current directory.

    Parameters
    ----------
        name : str
            source pdf name
    """
    import lhapdf  # pylint: disable=import-error, import-outside-toplevel

    print(f"install_pdf {name}")
    target = pathlib.Path(lhapdf.paths()[0])
    src = pathlib.Path(name)
    if not src.exists():
        raise FileExistsError(src)
    shutil.move(str(src), str(target))


def generate_block(xfxQ2, xgrid, Q2grid, pids):
    """
    Generate an LHAPDF data block from a callable

    Parameters
    ----------
        xfxQ2 : callable
            LHAPDF like callable
        Q2grid : list(float)
            Q grid
        pids : list(int)
            Flavors list
        xgrid : list(float)
            x grid

    Returns
    -------
        dict :
            PDF block
    """
    block = dict(Q2grid=Q2grid, pids=pids, xgrid=xgrid)
    data = []
    for x in xgrid:
        for Q2 in Q2grid:
            data.append(np.array([xfxQ2(pid, x, Q2) for pid in pids]))
    block["data"] = np.array(data)
    return block



# -*- coding: utf-8 -*-
"""
APFEL interface
"""
import time

import numpy as np
from banana.benchmark.external.apfel_utils import load_apfel

from eko import basis_rotation as br


def compute_apfel_data(
    theory, operators, pdf, skip_pdfs, rotate_to_evolution_basis=False
):

    """
    Run APFEL to compute operators.

    Parameters
    ----------
        theory : dict
            theory card
        operators : dict
            operators card
        pdf : lhapdf_type
            pdf
        skip_pdfs : list
            list of pdfs (pid or name) to skip
        rotate_to_evolution_basis: bool
            rotate to evolution basis

    Returns
    -------
        ref : dict
            output containing: target_xgrid, values
    """

    target_xgrid = operators["interpolation_xgrid"]
    pdf_name = pdf.set().name

    # Load apfel
    apf_start = time.perf_counter()
    if theory["ModEv"] in ["EXA", "perturbative-exact"]:
        theory["ModEv"] = "EXA"
    elif theory["ModEv"] in ["EXP", "decompose-expanded", "perturbative-expanded"]:
        theory["ModEv"] = "EXP"
    elif theory["ModEv"] in ["TRN", "ordered-truncated"]:
        theory["ModEv"] = "TRN"
    else:
        raise ValueError(f"Method {theory['ModEv']} is not recognized. ")
    apfel = load_apfel(theory, operators, pdf_name)

    # Truncated Epsilon
    # APFEL::SetEpsilonTruncation(1E-1);
    #
    # Set maximum scale
    # APFEL::SetQLimits(theory.Q0, theory.QM );
    #
    # if (theory.SIA)
    # {
    #   APFEL::SetPDFSet("kretzer");
    #   APFEL::SetTimeLikeEvolution(true);
    # }

    # Set APFEL interpolation grid
    #
    # apfel.SetNumberOfGrids(3)
    # apfel.SetGridParameters(1, 50, 3, 1e-5)
    # apfel.SetGridParameters(2, 50, 3, 2e-1)
    # apfel.SetGridParameters(3, 50, 3, 8e-1)

    # init evolution
    apfel.InitializeAPFEL()
    print(f"Loading APFEL took {(time.perf_counter() - apf_start)} s")

    # Run
    apf_tabs = {}
    for q2 in operators["Q2grid"]:

        apfel.EvolveAPFEL(theory["Q0"], np.sqrt(q2))
        print(f"Executing APFEL took {(time.perf_counter() - apf_start)} s")

        tab = {}
        for pid in br.flavor_basis_pids:

            if pid in skip_pdfs:
                continue

            # collect APFEL
            apf = []
            for x in target_xgrid:
                xf = apfel.xPDF(pid if pid != 21 else 0, x)
                # if pid == 4:
                #     print(pid,x,xf)
                apf.append(xf)
            tab[pid] = np.array(apf)

        # rotate if needed
        if rotate_to_evolution_basis:
            pdfs = np.array(
                [
                    tab[pid] if pid in tab else np.zeros(len(target_xgrid))
                    for pid in br.flavor_basis_pids
                ]
            )
            evol_pdf = br.rotate_flavor_to_evolution @ pdfs
            tab = dict(zip(br.evol_basis, evol_pdf))

        apf_tabs[q2] = tab

    ref = {
        "target_xgrid": target_xgrid,
        "values": apf_tabs,
    }

    return ref



# -*- coding: utf-8 -*-
"""
Collection of non-singlet EKOs.
"""

import numba as nb
import numpy as np

from .. import beta
from . import as4_evolution_integrals as as4_ei
from . import evolution_integrals as ei
from . import utils


@nb.njit(cache=True)
def U_vec(gamma_ns, nf, order):
    r"""Compute the elements of the non-singlet U vector.

    Parameters
    ----------
    gamma_ns : numpy.ndarray
        non-singlet anomalous dimension
    nf : int
        number of active flavors
    order : int
        perturbative order

    Returns
    -------
    np.ndarray
        U vector

    """
    U = np.zeros(order[0], dtype=np.complex_)
    beta0 = beta.beta_qcd((2, 0), nf)
    R0 = gamma_ns[0] / beta0
    U[0] = 1.0
    if order[0] >= 2:
        b1 = beta.b_qcd((3, 0), nf)
        R1 = gamma_ns[1] / beta0 - b1 * R0
        U[1] = R1
    if order[0] >= 3:
        b2 = beta.b_qcd((4, 0), nf)
        R2 = gamma_ns[2] / beta0 - b1 * R1 - b2 * R0
        U[2] = 0.5 * (R2 + U[1] * R1)
    if order[0] >= 4:
        b3 = beta.b_qcd((5, 0), nf)
        R3 = gamma_ns[3] / beta0 - b1 * R2 - b2 * R1 - b3 * R0
        U[3] = 1 / 3 * (R3 + R2 * U[1] + R1 * U[2])
    return U


@nb.njit(cache=True)
def lo_exact(gamma_ns, a1, a0, nf):
    """
    |LO| non-singlet exact EKO

    Parameters
    ----------
        gamma_ns : numpy.ndarray
            non-singlet anomalous dimensions
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_ns^0 : complex
            |LO| non-singlet exact EKO
    """
    return np.exp(gamma_ns[0] * ei.j00(a1, a0, nf))


@nb.njit(cache=True)
def nlo_exact(gamma_ns, a1, a0, nf):
    """
    |NLO| non-singlet exact EKO

    Parameters
    ----------
        gamma_ns : numpy.ndarray
            non-singlet anomalous dimensions
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_ns^1 : complex
            |NLO| non-singlet exact EKO
    """
    return np.exp(
        gamma_ns[0] * ei.j01_exact(a1, a0, nf) + gamma_ns[1] * ei.j11_exact(a1, a0, nf)
    )


@nb.njit(cache=True)
def nlo_expanded(gamma_ns, a1, a0, nf):
    """
    |NLO| non-singlet expanded EKO

    Parameters
    ----------
        gamma_ns : numpy.ndarray
            non-singlet anomalous dimensions
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_ns^1 : complex
            |NLO| non-singlet expanded EKO
    """
    return np.exp(
        gamma_ns[0] * ei.j01_expanded(a1, a0, nf)
        + gamma_ns[1] * ei.j11_expanded(a1, a0, nf)
    )


@nb.njit(cache=True)
def nnlo_exact(gamma_ns, a1, a0, nf):
    """
    |NNLO| non-singlet exact EKO

    Parameters
    ----------
        gamma_ns : numpy.ndarray
            non-singlet anomalous dimensions
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_ns^2 : complex
            |NNLO| non-singlet exact EKO
    """
    return np.exp(
        gamma_ns[0] * ei.j02_exact(a1, a0, nf)
        + gamma_ns[1] * ei.j12_exact(a1, a0, nf)
        + gamma_ns[2] * ei.j22_exact(a1, a0, nf)
    )


@nb.njit(cache=True)
def nnlo_expanded(gamma_ns, a1, a0, nf):
    """
    |NNLO| non-singlet expanded EKO

    Parameters
    ----------
        gamma_ns : numpy.ndarray
            non-singlet anomalous dimensions
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_ns^2 : complex
            |NNLO| non-singlet expanded EKO
    """
    return np.exp(
        gamma_ns[0] * ei.j02_expanded(a1, a0, nf)
        + gamma_ns[1] * ei.j12_expanded(a1, a0, nf)
        + gamma_ns[2] * ei.j22_expanded(a1, a0, nf)
    )


@nb.njit(cache=True)
def n3lo_expanded(gamma_ns, a1, a0, nf):
    """|N3LO| non-singlet expanded EKO

    Parameters
    ----------
    gamma_ns : numpy.ndarray
        non-singlet anomalous dimensions
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    nf : int
        number of active flavors

    Returns
    -------
    complex
        |N3LO| non-singlet expanded EKO

    """
    beta0 = beta.beta_qcd((2, 0), nf)
    b_list = [
        beta.b_qcd((3, 0), nf),
        beta.b_qcd((4, 0), nf),
        beta.b_qcd((5, 0), nf),
    ]
    j00 = ei.j00(a1, a0, nf)
    j13 = as4_ei.j13_expanded(a1, a0, beta0, b_list)
    j23 = as4_ei.j23_expanded(a1, a0, beta0, b_list)
    j33 = as4_ei.j33_expanded(a1, a0, beta0)
    return np.exp(
        gamma_ns[0] * as4_ei.j03_expanded(j00, j13, j23, j33, b_list)
        + gamma_ns[1] * j13
        + gamma_ns[2] * j23
        + gamma_ns[3] * j33
    )


@nb.njit(cache=True)
def n3lo_exact(gamma_ns, a1, a0, nf):
    """|N3LO| non-singlet exact EKO

    Parameters
    ----------
    gamma_ns : numpy.ndarray
        non-singlet anomalous dimensions
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    nf : int
        number of active flavors

    Returns
    -------
    complex
        |N3LO| non-singlet exact EKO

    """
    beta0 = beta.beta_qcd((2, 0), nf)
    b_list = [
        beta.b_qcd((3, 0), nf),
        beta.b_qcd((4, 0), nf),
        beta.b_qcd((5, 0), nf),
    ]
    roots = as4_ei.roots(b_list)
    j00 = ei.j00(a1, a0, nf)
    j13 = as4_ei.j13_exact(a1, a0, beta0, b_list, roots)
    j23 = as4_ei.j23_exact(a1, a0, beta0, b_list, roots)
    j33 = as4_ei.j33_exact(a1, a0, beta0, b_list, roots)
    return np.exp(
        gamma_ns[0] * as4_ei.j03_exact(j00, j13, j23, j33, b_list)
        + gamma_ns[1] * j13
        + gamma_ns[2] * j23
        + gamma_ns[3] * j33
    )


@nb.njit(cache=True)
def eko_ordered_truncated(gamma_ns, a1, a0, nf, order, ev_op_iterations):
    """|NLO|, |NNLO| or |N3LO| non-singlet ordered truncated EKO

    Parameters
    ----------
    gamma_ns : numpy.ndarray
        non-singlet anomalous dimensions
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    nf : int
        number of active flavors
    order : tuple(int,int)
        perturbative order
    ev_op_iterations : int
        number of evolution steps

    Returns
    -------
    complex
        non-singlet ordered truncated EKO

    """
    a_steps = utils.geomspace(a0, a1, 1 + ev_op_iterations)
    U = U_vec(gamma_ns, nf, order)
    e = 1.0
    al = a_steps[0]
    for ah in a_steps[1:]:
        e0 = lo_exact(gamma_ns, ah, al, nf)
        num, den = 0, 0
        for i in range(order[0]):
            num += U[i] * ah**i
            den += U[i] * al**i
        e *= e0 * num / den
        al = ah
    return e


@nb.njit(cache=True)
def eko_truncated(gamma_ns, a1, a0, nf, order, ev_op_iterations):
    """|NLO|, |NNLO| or |N3LO| non-singlet truncated EKO

    Parameters
    ----------
    gamma_ns : numpy.ndarray
        non-singlet anomalous dimensions
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    nf : int
        number of active flavors
    order : tuple(int,int)
        perturbative order
    ev_op_iterations : int
        number of evolution steps

    Returns
    -------
    complex
        non-singlet truncated EKO

    """
    a_steps = utils.geomspace(a0, a1, 1 + ev_op_iterations)
    U = U_vec(gamma_ns, nf, order)
    e = 1.0
    al = a_steps[0]
    fact = U[0]
    for ah in a_steps[1:]:
        e0 = lo_exact(gamma_ns, ah, al, nf)
        if order[0] >= 2:
            fact += U[1] * (ah - al)
        if order[0] >= 3:
            fact += +U[2] * ah**2 - ah * al * U[1] ** 2 + al**2 * (U[1] ** 2 - U[2])
        if order[0] >= 4:
            fact += (
                +(ah**3) * U[3]
                - ah**2 * al * U[2] * U[1]
                + ah * al**2 * U[1] * (U[1] ** 2 - U[2])
                - al**3 * (U[1] ** 3 - 2 * U[1] * U[2] + U[3])
            )
        e *= e0 * fact
        al = ah
    return e


@nb.njit(cache=True)
def dispatcher(
    order, method, gamma_ns, a1, a0, nf, ev_op_iterations
):  # pylint: disable=too-many-return-statements
    """Determine used kernel and call it.

    In LO we always use the exact solution.

    Parameters
    ----------
    order : tuple(int,int)
        perturbation order
    method : str
        method
    gamma_ns : numpy.ndarray
        non-singlet anomalous dimensions
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    nf : int
        number of active flavors
    ev_op_iterations : int
        number of evolution steps

    Returns
    -------
    complex
        non-singlet EKO

    """
    # use always exact in LO
    if order[0] == 1:
        return lo_exact(gamma_ns, a1, a0, nf)
    if method == "ordered-truncated":
        return eko_ordered_truncated(gamma_ns, a1, a0, nf, order, ev_op_iterations)
    if method == "truncated":
        return eko_truncated(gamma_ns, a1, a0, nf, order, ev_op_iterations)

    # NLO
    if order[0] == 2:
        if method in [
            "iterate-expanded",
            "decompose-expanded",
            "perturbative-expanded",
        ]:
            return nlo_expanded(gamma_ns, a1, a0, nf)
        # if method in ["iterate-exact", "decompose-exact", "perturbative-exact"]:
        return nlo_exact(gamma_ns, a1, a0, nf)
    # NNLO
    if order[0] == 3:
        if method in [
            "iterate-expanded",
            "decompose-expanded",
            "perturbative-expanded",
        ]:
            return nnlo_expanded(gamma_ns, a1, a0, nf)
        return nnlo_exact(gamma_ns, a1, a0, nf)
    # N3LO
    if order[0] == 4:
        if method in [
            "iterate-expanded",
            "decompose-expanded",
            "perturbative-expanded",
        ]:
            return n3lo_expanded(gamma_ns, a1, a0, nf)
        return n3lo_exact(gamma_ns, a1, a0, nf)
    raise NotImplementedError("Selected order is not implemented")



# -*- coding: utf-8 -*-
"""
Plotting tools to show the evolved PDF and the computed operators
"""
import io
import pprint

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.cm import get_cmap
from matplotlib.colors import LogNorm

from eko import basis_rotation as br


def input_figure(theory, ops, pdf_name=None):
    """
    Pretty-prints the setup to a figure

    Parameters
    ----------
        theory : dict
            theory card
        ops : dict
            operator card
        pdf_name : str
            PDF name

    Returns
    -------
        firstPage : matplotlib.pyplot.Figure
            figure
    """
    firstPage = plt.figure(figsize=(25, 20))
    # theory
    firstPage.text(0.05, 0.97, "Theory:", size=20, ha="left", va="top")
    str_stream = io.StringIO()
    th_copy = theory.copy()
    th_copy.update({"hash": theory["hash"][:7]})
    pprint.pprint(th_copy, stream=str_stream, width=50)
    firstPage.text(0.05, 0.92, str_stream.getvalue(), size=14, ha="left", va="top")
    # operators
    firstPage.text(0.55, 0.87, "Operators:", size=20, ha="left", va="top")
    str_stream = io.StringIO()
    ops_copy = ops.copy()
    ops_copy.update({"hash": ops["hash"][:7]})
    pprint.pprint(ops_copy, stream=str_stream, width=50)
    firstPage.text(0.55, 0.82, str_stream.getvalue(), size=14, ha="left", va="top")
    # pdf
    if pdf_name is not None:
        firstPage.text(0.55, 0.97, "source pdf:", size=20, ha="left", va="top")
        firstPage.text(0.55, 0.92, pdf_name, size=14, ha="left", va="top")
        firstPage.tight_layout()
    return firstPage


def plot_dist(x, y, yerr, yref, title=None, oMx_min=1e-2, oMx_max=0.5):
    """
    Compare two distributions.

    Generates a plot with 3 areas:

    - small x, i.e. log(x) as abscissa
    - linear x, i.e. with id(x) as abscissa
    - large x, i.e. with log(1-x) as abscissa


    Parameters
    ----------
        x : numpy.ndarray
            list of abscisses
        y : numpy.ndarray
            computed list of ordinates
        yerr : numpy.ndarray
            list of ordinate errors
        yref : numpy.ndarray
            reference list of ordinates

    Other Parameters
    ----------------
        title : str, optional
            additional overall title
        oMx_min : float
            maximum value for the large x region, i.e. 1-x > 1 - `oMx_min`
        oMx_max : float
            minimum value for the large x region, i.e. 1 - `oMx_max` > 1-x

    Returns
    -------
        fig : matplotlib.pyplot.figure
            created figure
    """
    np.seterr(divide="ignore", invalid="ignore")
    fig = plt.figure(figsize=(15, 5))
    fig.subplots_adjust(hspace=0.05)
    if title is not None:
        fig.suptitle(title)
    # small x
    ax1 = plt.subplot(2, 3, 1)
    plt.setp(ax1.get_xticklabels(), visible=False)
    plt.title("small x")
    ax1.set_xscale("log", nonpositive="clip")
    ax1.set_yscale("log", nonpositive="clip")
    plt.errorbar(x, y, yerr=yerr, fmt="o")
    plt.loglog(x, yref, "x")
    ax1b = plt.subplot(2, 3, 4, sharex=ax1)
    ax1b.set_xscale("log", nonpositive="clip")
    # ax1b.set_yscale("log", nonpositive="clip")
    # plt.errorbar(x, np.abs((y - yref) / yref), yerr=np.abs(yerr / yref), fmt="s")
    y_rel_diff = (y - yref) / yref
    ax1b.set_yscale("symlog", linthresh=np.min(np.abs(y_rel_diff)), linscale=0.1)
    plt.errorbar(x, y_rel_diff, yerr=np.abs(yerr / yref), fmt="s")
    plt.axhline(0, linestyle="--", color="grey")
    plt.xlabel("x")
    # linear x
    ax2 = plt.subplot(2, 3, 2)
    plt.setp(ax2.get_xticklabels(), visible=False)
    plt.title("linear x")
    plt.errorbar(x, y, yerr=yerr, fmt="o")
    plt.plot(x, yref, "x")
    ax2b = plt.subplot(2, 3, 5, sharex=ax2)
    ax2b.set_yscale("log", nonpositive="clip")
    plt.errorbar(x, np.abs((y - yref) / yref), yerr=np.abs(yerr / yref), fmt="s")
    plt.xlabel("x")
    # large x
    ax3 = plt.subplot(2, 3, 3)
    ax3.set_xscale("log", nonpositive="clip")
    ax3.set_yscale("log", nonpositive="clip")
    x = np.array(x)
    oMx = 1.0 - x
    plt.setp(ax3.get_xticklabels(), visible=False)
    ax3.set_xlim(oMx_min, oMx_max)
    plt.title("large x, i.e. small (1-x)")
    plt.errorbar(oMx, y, yerr=yerr, fmt="o")
    plt.loglog(oMx, yref, "x")
    ax3b = plt.subplot(2, 3, 6, sharex=ax3)
    ax3b.set_xscale("log", nonpositive="clip")
    ax3b.set_yscale("log", nonpositive="clip")
    plt.errorbar(oMx, np.abs((y - yref) / yref), yerr=np.abs(yerr / yref), fmt="s")
    plt.xlabel("1-x")
    return fig


def plot_operator(var_name, op, op_err, log_operator=True, abs_operator=True):
    """
    Plot a single operator as heat map.

    Parameters
    ----------
        ret : dict
            DGLAP result
        var_name : str
            operator name
        log_operator : bool, optional
            plot on logarithmic scale
        abs_operator : bool, optional
            plot absolute value (instead of true value)

    Returns
    -------
        fig : matplotlib.pyplot.figure
            created figure
    """
    # get
    # op = ret["operators"][var_name]
    # op_err = ret["operator_errors"][var_name]

    # empty?
    thre = 1e-8
    if np.max(np.abs(op)) <= thre:
        return None

    fig = plt.figure(figsize=(25, 5))
    fig.suptitle(var_name)

    # TODO fix File "/usr/lib/python3/dist-packages/matplotlib/colors.py",
    # line 1181, in _check_vmin_vmax
    # raise ValueError("minvalue must be positive")

    # Settings
    colormap = "coolwarm"
    ax = plt.subplot(1, 3, 1)
    if abs_operator:
        plt.title("|operator|")
        colormap = "inferno"
        op = np.abs(op)
    else:
        plt.title("operator")

    norm = LogNorm() if log_operator else None
    cmap = get_cmap(colormap)
    # Plotting operator
    im = plt.imshow(op, cmap=cmap, norm=norm, aspect="auto")
    plt.colorbar(im, ax=ax, fraction=0.034, pad=0.04)
    # Plotting operator error
    ax = plt.subplot(1, 3, 2)
    plt.title("operator_error")
    im = plt.imshow(op_err, norm=norm, aspect="auto")
    plt.colorbar(im, ax=ax, fraction=0.034, pad=0.04)
    # Plotting ratio
    ax = plt.subplot(1, 3, 3)
    plt.title("|error/value|")
    # Avoid the numpy warning
    old_settings = np.seterr(divide="ignore", invalid="ignore")
    err_to_val = np.abs(np.array(op_err) / np.array(op))
    np.seterr(**old_settings)

    im = plt.imshow(err_to_val, norm=norm, aspect="auto")
    plt.colorbar(im, ax=ax, fraction=0.034, pad=0.04)
    return fig


def save_operators_to_pdf(path, theory, ops, me, skip_pdfs, change_lab=False):
    """
        Output all operator heatmaps to PDF.

    Parameters
    ----------
        path : str
            path to the plot
        theory : dict
            theory card
        ops : dict
            operator card
        me : eko.output.Output
            DGLAP result
        skip_pdfs : list
            PDF to skip
        change_lab : bool
            set whether to rename the labels
    """
    ops_names = list(me["targetpids"])
    ops_id = f"o{ops['hash'][:6]}_t{theory['hash'][:6]}"
    path = f"{path}/{ops_id}.pdf"
    print(f"Plotting operators plots to {path}")
    with PdfPages(path) as pp:
        # print setup
        firstPage = input_figure(theory, ops)
        pp.savefig()
        plt.close(firstPage)

        # plot the operators
        # it's necessary to reshuffle the eko output
        for q2 in me["Q2grid"]:
            results = me["Q2grid"][q2]["operators"]
            errors = me["Q2grid"][q2]["operator_errors"]

            # loop on pids
            for label_out, res, res_err in zip(ops_names, results, errors):
                if label_out in skip_pdfs:
                    continue
                new_op = {}
                new_op_err = {}
                # loop on xgrid point
                for j in range(len(me["interpolation_xgrid"])):
                    # loop on pid in
                    for label_in, val, val_err in zip(ops_names, res[j], res_err[j]):
                        if label_in in skip_pdfs:
                            continue
                        if label_in not in new_op:
                            new_op[label_in] = []
                            new_op_err[label_in] = []
                        new_op[label_in].append(val)
                        new_op_err[label_in].append(val_err)

                for label_in in ops_names:
                    if label_in in skip_pdfs:
                        continue
                    lab_in = label_in
                    lab_out = label_out
                    if change_lab:
                        index_in = br.evol_basis_pids.index(label_in)
                        index_out = br.evol_basis_pids.index(label_out)
                        lab_in = br.evol_basis[index_in]
                        lab_out = br.evol_basis[index_out]
                    try:
                        fig = plot_operator(
                            f"Operator ({lab_in};{lab_out}) µ_F^2 = {q2} GeV^2",
                            new_op[label_in],
                            new_op_err[label_in],
                        )
                        if fig is not None:
                            pp.savefig()
                    finally:
                        if fig:
                            plt.close(fig)



# -*- coding: utf-8 -*-
"""Some utility functions"""

import numba as nb
import numpy as np


@nb.njit(cache=True)
def geomspace(start, end, num):
    """
    Numba port of :func:`numpy.geomspace`.

    Parameters
    ----------
        start : float
            initial value
        end : float
            final value
        num : int
            steps

    Returns
    -------
        geomspace : numpy.ndarray
            logarithmic spaced list between `start` and `end`
    """
    return np.exp(np.linspace(np.log(start), np.log(end), num))



# -*- coding: utf-8 -*-
"""This subpackage contains the |N3LO| Altarelli-Parisi splitting kernels.

For further documentation see :doc:`N3LO anomalous dimensions <../../../theory/N3LO_ad>`

"""
import numba as nb
import numpy as np

from .ggg import gamma_gg
from .ggq import gamma_gq
from .gnsm import gamma_nsm
from .gnsp import gamma_nsp
from .gnsv import gamma_nsv
from .gps import gamma_ps
from .gqg import gamma_qg


@nb.njit(cache=True)
def gamma_singlet(N, nf, sx):
    r"""Computes the |N3LO| singlet anomalous dimension matrix

      .. math::
          \gamma_S^{(3)} = \left(\begin{array}{cc}
          \gamma_{qq}^{(3)} & \gamma_{qg}^{(3)}\\
          \gamma_{gq}^{(3)} & \gamma_{gg}^{(3)}
          \end{array}\right)

    Parameters
    ----------
    N : complex
      Mellin moment
    nf : int
      Number of active flavors
    sx : list
      harmonic sums cache

    Returns
    -------
    numpy.ndarray
        |N3LO| singlet anomalous dimension matrix
        :math:`\gamma_{S}^{(3)}(N)`

    See Also
    --------
    gamma_nsp : :math:`\gamma_{ns,+}^{(3)}`
    gamma_ps : :math:`\gamma_{ps}^{(3)}`
    gamma_qg : :math:`\gamma_{qg}^{(3)}`
    gamma_gq : :math:`\gamma_{gq}^{(3)}`
    gamma_gg : :math:`\gamma_{gg}^{(3)}`

    """
    gamma_qq = gamma_nsp(N, nf, sx) + gamma_ps(N, nf, sx)
    gamma_S_0 = np.array(
        [[gamma_qq, gamma_qg(N, nf, sx)], [gamma_gq(N, nf, sx), gamma_gg(N, nf, sx)]],
        np.complex_,
    )
    return gamma_S_0



# -*- coding: utf-8 -*-
"""
This file contains the O(as1aem1) Altarelli-Parisi splitting kernels.
"""

import numba as nb

from .. import constants, harmonics
from ..harmonics.constants import zeta2, zeta3


@nb.njit(cache=True)
def gamma_phq(N, sx):
    """Computes the O(as1aem1) photon-quark anomalous dimension

    Implements Eq. (36) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
        N : complex
            Mellin moment
        sx : np array
            List of harmonic sums

    Returns
    -------
      gamma_phq : complex
        O(as1aem1) photon-quark anomalous dimension :math:`\\gamma_{\\gamma q}^{(1,1)}(N)`

    """
    S1 = sx[0]
    S2 = sx[1]
    tmp_const = (
        2
        * (-4 - 12 * N - N**2 + 28 * N**3 + 43 * N**4 + 30 * N**5 + 12 * N**6)
        / ((-1 + N) * N**3 * (1 + N) ** 3)
    )
    tmp_S1 = (
        -4
        * (10 + 27 * N + 25 * N**2 + 13 * N**3 + 5 * N**4)
        / ((-1 + N) * N * (1 + N) ** 3)
    )
    tmp_S12 = 4 * (2 + N + N**2) / ((-1 + N) * N * (1 + N))
    tmp_S2 = 4 * (2 + N + N**2) / ((-1 + N) * N * (1 + N))

    return constants.CF * (tmp_const + tmp_S1 * S1 + tmp_S12 * S1**2 + tmp_S2 * S2)


@nb.njit(cache=True)
def gamma_qph(N, nf, sx):
    """Computes the O(as1aem1) quark-photon anomalous dimension

    Implements Eq. (26) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
        N : complex
            Mellin moment
        nf : int
            Number of active flavors
        sx : np array
            List of harmonic sums

    Returns
    -------
      gamma_qph : complex
        O(as1aem1) quark-photon anomalous dimension :math:`\\gamma_{q \\gamma}^{(1,1)}(N)`

    """
    S1 = sx[0]
    S2 = sx[1]
    tmp_const = (
        -2
        * (
            4
            + 8 * N
            + 25 * N**2
            + 51 * N**3
            + 36 * N**4
            + 15 * N**5
            + 5 * N**6
        )
        / (N**3 * (1 + N) ** 3 * (2 + N))
    )
    tmp_S1 = 8 / N**2
    tmp_S12 = -4 * (2 + N + N**2) / (N * (1 + N) * (2 + N))
    tmp_S2 = 4 * (2 + N + N**2) / (N * (1 + N) * (2 + N))
    return (
        2
        * nf
        * constants.CA
        * constants.CF
        * (tmp_const + tmp_S1 * S1 + tmp_S12 * S1**2 + tmp_S2 * S2)
    )


@nb.njit(cache=True)
def gamma_gph(N):
    """Computes the O(as1aem1) gluon-photon anomalous dimension

    Implements Eq. (27) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
      N : complex
        Mellin moment

    Returns
    -------
      gamma_qph : complex
        O(as1aem1) gluon-photon anomalous dimension :math:`\\gamma_{g \\gamma}^{(1,1)}(N)`

    """
    return (
        constants.CF
        * constants.CA
        * (8 * (-4 + N * (-4 + N * (-5 + N * (-10 + N + 2 * N**2 * (2 + N))))))
        / (N**3 * (1 + N) ** 3 * (-2 + N + N**2))
    )


@nb.njit(cache=True)
def gamma_phg(N):
    """Computes the O(as1aem1) photon-gluon anomalous dimension

    Implements Eq. (30) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
      N : complex
        Mellin moment

    Returns
    -------
      gamma_qph : complex
        O(as1aem1) photon-gluon anomalous dimension :math:`\\gamma_{\\gamma g}^{(1,1)}(N)`

    """
    return constants.TR / constants.CF / constants.CA * gamma_gph(N)


@nb.njit(cache=True)
def gamma_qg(N, nf, sx):
    """Computes the O(as1aem1) quark-gluon singlet anomalous dimension.

    Implements Eq. (29) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
        N : complex
            Mellin moment
        nf : int
            Number of active flavors
        sx : np array
            List of harmonic sums

    Returns
    -------
        gamma_qg : complex
            O(as1aem1) quark-gluon singlet anomalous dimension
            :math:`\\gamma_{qg}^{(1,1)}(N)`

    """
    return constants.TR / constants.CF / constants.CA * gamma_qph(N, nf, sx)


@nb.njit(cache=True)
def gamma_gq(N, sx):
    """Computes the O(as1aem1) gluon-quark singlet anomalous dimension.

    Implements Eq. (35) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
        N : complex
            Mellin moment
        sx : np array
            List of harmonic sums

    Returns
    -------
        gamma_gq : complex
            O(as1aem1) gluon-quark singlet anomalous dimension
            :math:`\\gamma_{gq}^{(1,1)}(N)`

    """
    return gamma_phq(N, sx)


@nb.njit(cache=True)
def gamma_phph(nf):
    """Computes the O(as1aem1) photon-photon singlet anomalous dimension.

    Implements Eq. (28) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
        nf : int
            Number of active flavors

    Returns
    -------
        gamma_gg : complex
            O(as1aem1) photon-photon singlet anomalous dimension
            :math:`\\gamma_{\\gamma \\gamma}^{(1,1)}(N)`

    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    return 4 * constants.CF * constants.CA * (nu * constants.eu2 + nd * constants.ed2)


@nb.njit(cache=True)
def gamma_gg():
    """Computes the O(as1aem1) gluon-gluon singlet anomalous dimension.

    Implements Eq. (31) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------

    Returns
    -------
        gamma_gg : complex
            O(as1aem1) gluon-gluon singlet anomalous dimension
            :math:`\\gamma_{gg}^{(1,1)}(N)`

    """
    return 4 * constants.TR


@nb.njit(cache=True)
def gamma_nsp(N, sx):
    """Computes the O(as1aem1) singlet-like non-singlet anomalous dimension.

    Implements sum of Eqs. (33-34) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
        N : complex
            Mellin moment
        sx : np array
            List of harmonic sums

    Returns
    -------
        gamma_nsp : complex
            O(as1aem1) singlet-like non-singlet anomalous dimension
            :math:`\\gamma_{ns,+}^{(1)}(N)`

    """
    S1 = sx[0]
    S2 = sx[1]
    S3 = sx[2]
    S1h = harmonics.S1(N / 2)
    S2h = harmonics.S2(N / 2)
    S3h = harmonics.S3(N / 2)
    S1p1h = harmonics.S1((N + 1.0) / 2)
    S2p1h = harmonics.S2((N + 1) / 2)
    S3p1h = harmonics.S3((N + 1) / 2)
    g3N = harmonics.g_functions.mellin_g3(N, S1)
    S1p2 = harmonics.polygamma.recursive_harmonic_sum(S1, N, 2, 1)
    g3Np2 = harmonics.g_functions.mellin_g3(N + 2, S1p2)
    result = (
        +32 * zeta2 * S1h
        - 32 * zeta2 * S1p1h
        + 8.0 / (N + N**2) * S2h
        - 4 * S3h
        + (24 + 16 / (N + N**2)) * S2
        - 32 * S3
        - 8.0 / (N + N**2) * S2p1h
        + S1
        * (
            +16 * (3 / N**2 - 3 / (1 + N) ** 2 + 2 * zeta2)
            - 16 * S2h
            - 32 * S2
            + 16 * S2p1h
        )
        + (
            -8
            + N
            * (
                -32
                + N * (-8 - 3 * N * (3 + N) * (3 + N**2) - 48 * (1 + N) ** 2 * zeta2)
            )
        )
        / (N**3 * (1 + N) ** 3)
        + 32 * (g3N + g3Np2)
        + 4 * S3p1h
        - 16 * zeta3
    )
    return constants.CF * result


@nb.njit(cache=True)
def gamma_nsm(N, sx):
    """Computes the O(as1aem1) valence-like non-singlet anomalous dimension.

    Implements difference between Eqs. (33-34) of :cite:`deFlorian:2015ujt`.

    Parameters
    ----------
        N : complex
            Mellin moment
        sx : np array
            List of harmonic sums

    Returns
    -------
        gamma_nsm : complex
            O(as1aem1) singlet-like non-singlet anomalous dimension
            :math:`\\gamma_{ns,-}^{(1,1)}(N)`

    """
    S1 = sx[0]
    S2 = sx[1]
    S3 = sx[2]
    S1h = harmonics.S1(N / 2)
    S2h = harmonics.S2(N / 2)
    S3h = harmonics.S3(N / 2)
    S1p1h = harmonics.S1((N + 1.0) / 2)
    S2p1h = harmonics.S2((N + 1) / 2)
    S3p1h = harmonics.S3((N + 1) / 2)
    g3N = harmonics.g_functions.mellin_g3(N, S1)
    S1p2 = harmonics.polygamma.recursive_harmonic_sum(S1, N, 2, 1)
    g3Np2 = harmonics.g_functions.mellin_g3(N + 2, S1p2)

    result = (
        -32.0 * zeta2 * S1h
        - 8.0 / (N + N**2) * S2h
        + (24 + 16 / (N + N**2)) * S2
        + 8.0 / (N + N**2) * S2p1h
        + S1
        * (
            16 * (-1 / N**2 + 1 / (1 + N) ** 2 + 2 * zeta2)
            + 16 * S2h
            - 32 * S2
            - 16 * S2p1h
        )
        + (
            72
            + N
            * (
                96
                - 3 * N * (8 + 3 * N * (3 + N) * (3 + N**2))
                + 48 * N * (1 + N) ** 2 * zeta2
            )
        )
        / (3.0 * N**3 * (1 + N) ** 3)
        - 32 * (g3N + g3Np2)
        + 32.0 * zeta2 * S1p1h
        + 4 * S3h
        - 32 * S3
        - 4 * S3p1h
        - 16 * zeta3
    )
    return constants.CF * result



# -*- coding: utf-8 -*-
"""
Pegasus interface
"""
import numpy as np

from eko import basis_rotation as br


def compute_pegasus_data(theory, operators, skip_pdfs, rotate_to_evolution_basis=False):
    """
    Run Pegasus to compute operators.

    Parameters
    ----------
        theory : dict
            theory card
        operators : dict
            operators card
        skip_pdfs : list
            list of pdfs (pid or name) to skip
        rotate_to_evolution_basis: bool
            rotate to evolution basis

    Returns
    -------
        ref : dict
            output containing: target_xgrid, values
    """
    import pegasus  # pylint:disable=import-error,import-outside-toplevel

    target_xgrid = operators["interpolation_xgrid"]

    # init pegasus
    nf = theory["NfFF"]

    if theory["ModEv"] in ["EXA", "perturbative-exact"]:
        imodev = 1
    elif theory["ModEv"] in ["EXP", "decompose-expanded", "perturbative-expanded"]:
        imodev = 2
    elif theory["ModEv"] in ["ordered-truncated"]:
        imodev = 3
    elif theory["ModEv"] in ["TRN"]:
        imodev = 4  # 4 is random - just not 1-3
    else:
        raise ValueError(f"Method {theory['ModEv']} is not recognized. ")

    ivfns = 0 if theory["FNS"] == "FFNS" else 1

    if theory["Q0"] != theory["Qref"]:
        raise ValueError("Initial scale Q0 must be equal to Qref in Pegasus.")

    pegasus.initevol(
        imodev, theory["PTO"], ivfns, nf, theory["fact_to_ren_scale_ratio"] ** 2
    )
    pegasus.initinp(
        theory["alphas"],
        theory["Qref"] ** 2,
        (theory["kcThr"] * theory["mc"]) ** 2,
        (theory["kbThr"] * theory["mb"]) ** 2,
        (theory["ktThr"] * theory["mt"]) ** 2,
    )

    # better return always the flavor basis and then rotate
    # if rotate_to_evolution_basis:
    #     ipstd = 0
    #     labels = list(br.evol_basis)
    # else:
    #     ipstd = 1

    # photon pdf is not in pagsus output
    labels = list(br.flavor_basis_pids)
    labels.remove(22)
    # swap u, d and ubar and dbar
    labels[4], labels[5] = labels[5], labels[4]
    labels[-6], labels[-5] = labels[-5], labels[-6]

    # run pegaus
    out_tabs = {}
    for q2 in operators["Q2grid"]:

        tab = {}
        for x in target_xgrid:
            # last two numbers are the min and max pid to calculate,
            # keep everthing for simplicity.
            xf, _ = pegasus.xparton(x, q2, -6, 6)
            temp = dict(zip(labels, xf))
            for pid in labels:
                if pid in skip_pdfs:
                    continue
                if pid not in tab:
                    tab[pid] = []
                tab[pid].append(temp[pid])

        # rotate if needed
        if rotate_to_evolution_basis:
            pdfs = np.array(
                [
                    tab[pid] if pid in tab else np.zeros(len(target_xgrid))
                    for pid in br.flavor_basis_pids
                ]
            )
            evol_pdf = br.rotate_flavor_to_evolution @ pdfs
            tab = dict(zip(br.evol_basis, evol_pdf))

        out_tabs[q2] = tab

    ref = {"target_xgrid": target_xgrid, "values": out_tabs}

    return ref
